<?php

namespace EventPortal\Domain\Form;

use Closure;
use EventPortal\Domain\Component\Section;
use EventPortal\Domain\Util\ValidationRules;
use EventPortal\Exceptions\NoSaveHandlers;
use Illuminate\Database\Eloquent\Model;
use Illuminate\Pipeline\Pipeline;
use Illuminate\Support\Collection;
use Illuminate\Support\Facades\Validator;

class Form
{
    /** @var string */
    public $name;

    /** @var array */
    public $blocks = [];

    /** @var array */
    public $actions = [];

    /** @var Model */
    private $model;

    /** @var array */
    private $context = [];

    /** @var array */
    private $saveHandlers;

    public function __construct(string $name)
    {
        $this->name = $name;
    }

    /**
     * @param callable|string $fields
     */
    public function addFields($fields): self
    {
        $this->blocks[] = $fields;

        return $this;
    }

    public function addActions(callable $actions): self
    {
        $this->actions = $actions;

        return $this;
    }

    /**
     * @param mixed ...$models
     */
    public function context(...$models): self
    {
        $keyedModels = [];

        foreach ($models as $model) {
            $keyedModels[$model::KEY] = $model;
        }

        $this->context = array_merge($keyedModels, [$this->model::KEY => $this->model]);

        return $this;
    }

    public function model(Model $model): self
    {
        $this->model = $model;

        return $this;
    }

    public function validate(array $input): self
    {
        $this->resolveWithoutHydration();

        Validator::make($input, ValidationRules::for($this->blocks))->validate();

        return $this;
    }

    public function render(): array
    {
        $this->resolve();

        return [
            'actions' => $this->actions,
            'content' => $this->blocks,
        ];
    }

    public function addSaveHandler(callable $saveHandler): self
    {
        $this->saveHandlers[] = $saveHandler;

        return $this;
    }

    /**
     * @return mixed
     */
    public function save(array $input)
    {
        throw_if($this->saveHandlers === [], NoSaveHandlers::class);

        $payload = Payload::make($input)
            ->withModel($this->model)
            ->withContext($this->context);

        return app(Pipeline::class)
            ->send($payload)
            ->through($this->saveHandlers)
            ->thenReturn();
    }

    private function resolveWithoutHydration(): void
    {
        $this->resolve(false);
    }

    private function resolve(bool $hydrate = true): void
    {
        $this->resolveActions();
        $this->resolveBlocks($hydrate);
    }

    private function resolveActions(): void
    {
        if ($this->actions === []) {
            return;
        }

        $parameters = array_merge($this->context, [$this->model::KEY => $this->model]);

        $this->actions = app()->call($this->actions, $parameters);
    }

    private function resolveBlocks(bool $hydrate): void
    {
        $this->blocks = collect($this->blocks)
            ->map(function ($field) {
                $callback = $field instanceof Closure ? $field : app($field);

                return app()->call($callback);
            })
            ->flatten()
            ->map(function ($section) {
                if ($section instanceof Section) {
                    $section->children = $section->flatChildren();

                    return $section;
                }

                return Section::make()->fields([$section]);
            })
            ->when($hydrate, function (Collection $sections) {
                return $sections->each(function (Section $section): void {
                    foreach ($section->flatChildren() as $input) {
                        $input->value($this->getValue($input));
                    }
                });
            })
            ->toArray();
    }

    /**
     * @param mixed $input
     * @return mixed|null
     */
    private function getValue($input)
    {
        $name = $input->name;

        if ($input->isNotScoped()) {
            return optional($this->model)->$name;
        }

        return object_get($this->model, $input->scopeKey());
    }
}
